class TestModel:
    def test_example(self, example_fragments, example_clustering):
        pass
        # fragments = []
        # clustering = []
        # fold = []
        # model = []
        # result = model.get_prediction(
        #     fragments, clustering, cell_ixs=fold["cells_test"]
        # )
